/*Copyright 2011, 2013 Alexandros Chortaras

 This file is part of Rapid.

 Rapid is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Rapid is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Rapid.  If not, see <http://www.gnu.org/licenses/>.*/

package edu.ntua.ivml.qa.algorithm.rapid;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Set;

import edu.ntua.ivml.common.lp.Atom;
import edu.ntua.ivml.common.lp.AtomArray;
import edu.ntua.ivml.common.lp.Clause;
import edu.ntua.ivml.common.lp.Function;
import edu.ntua.ivml.common.lp.FunctionalTerm;
import edu.ntua.ivml.common.lp.Substitution;
import edu.ntua.ivml.common.lp.Term;
import edu.ntua.ivml.common.lp.Variable;
import edu.ntua.ivml.qa.algorithm.rapid.dllite.DLAtomUnfolding;
import edu.ntua.ivml.qa.lp.theory.FCTheory;


public abstract class AtomUnfolder {

	protected FCTheory program;
	
	protected AtomUnfolder(FCTheory program) {
		this.program = program;
	}
	
	public abstract AtomUnfoldingResult getUnfolding(Atom atom, Collection<? extends Term> boundTerms, Atom head, FlattenerExtra extra);

	public abstract AtomUnfoldingResult getUnfolding(Set<Atom> atom, Collection<? extends Term> boundTerms, Atom head, FlattenerExtra extra);

	public abstract AtomUnfoldingResult getShallowUnfolding(Atom atom, Collection<? extends Term> boundTerms);

	public boolean addMode = false;
	
	public static ArrayList<Atom> functionCollect(Atom atom, Clause sideClause, Set<Function>[] argFunctions) {
		if (!sideClause.isFunctionFree()) {
			Term[] sideTuple = sideClause.getHead().getArguments().getTerms();
			Term[] atomTuple = atom.getArguments().getTerms();
			
			Substitution s = new Substitution();
			
			if (atom.isRole()) {
				if (sideTuple[0].isFunctionalTerm() && !atomTuple[0].hasUnboundPrefix()) {
					if (atomTuple[0].equals(Rapid.dummyVars.get(0))) {
						argFunctions[0].add(((FunctionalTerm)sideTuple[0]).getFunction());
					} else {
						argFunctions[1].add(((FunctionalTerm)sideTuple[0]).getFunction());
					}
					s.put((Variable)sideTuple[1], Rapid.dummyVars.get(0));
					return AtomArray.apply(sideClause.getBody(), s);
				} else if (sideTuple[1].isFunctionalTerm() && !atomTuple[1].hasUnboundPrefix()) {
					if (atomTuple[1].equals(Rapid.dummyVars.get(1))) {
						argFunctions[1].add(((FunctionalTerm)sideTuple[1]).getFunction());
					} else {
						argFunctions[0].add(((FunctionalTerm)sideTuple[1]).getFunction());
					}
					s.put((Variable)sideTuple[0], Rapid.dummyVars.get(0));
					return AtomArray.apply(sideClause.getBody(), s);
				}
			} else {
				if (!atomTuple[0].hasUnboundPrefix()) {
					if (atomTuple[0].equals(Rapid.dummyVars.get(0))) {
	   					argFunctions[0].add(((FunctionalTerm)sideTuple[0]).getFunction());
	   				} else {
	   					argFunctions[1].add(((FunctionalTerm)sideTuple[0]).getFunction());
	   				}
					
					s.put(sideTuple[0].getVariables().iterator().next(), Rapid.dummyVars.get(0));
					return AtomArray.apply(sideClause.getBody(), s);
				}
			}
		} 
		
		return null;
	}

	public abstract Atom[] datalogComputeUnfolding(Atom atom, Collection<? extends Term> boundTerms, Atom head, FlattenerExtra extra);
	
	
	protected boolean preinit = false;
	
    public void precompute(Collection<Clause> atomClauses) {
    	preinit = true;
    	for (Clause c : atomClauses) {
    		getUnfolding(c.getBodyAtomAt(0), c.getHead().getVariables(), c.getHead(), new FlattenerExtra());
    	}
    }
    
    public boolean isPrecomputed() {
    	return preinit;
    }

}
