/*Copyright 2011, 2013 Alexandros Chortaras

 This file is part of Rapid.

 Rapid is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Rapid is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Rapid.  If not, see <http://www.gnu.org/licenses/>.*/

package edu.ntua.ivml.qa.lp.theory;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import edu.ntua.ivml.common.lp.Atom;
import edu.ntua.ivml.common.lp.Clause;
import edu.ntua.ivml.common.lp.Function;
import edu.ntua.ivml.common.lp.FunctionalTerm;
import edu.ntua.ivml.common.lp.Predicate;
import edu.ntua.ivml.common.lp.Term;


public class LogicProgram {

	protected ArrayList<Clause> rules;
	protected Map<Predicate, ArrayList<Clause>> ruleIndex;
	protected Map<Predicate, ArrayList<Clause>> iruleIndex;
	
	protected Map<Function, Set<Clause>> fIndex;
	
	public LogicProgram() {
		rules = new ArrayList<Clause>();
		ruleIndex = new HashMap<Predicate, ArrayList<Clause>>();
		iruleIndex = new HashMap<Predicate, ArrayList<Clause>>();
		
		fIndex = new HashMap<>();
	}
	
	public LogicProgram(Collection<Clause> set) {		
		this();
		
		for (Clause c : set) {
			addClause(c);
		}
	}
	
	public int size() {
		return rules.size();
	}
	
	public void addClause(Clause r) {
		
		if (rules.add(r)) {
			
			ArrayList<Clause> clauses = ruleIndex.get(r.getHead().getPredicate());
			if (clauses == null) {
				clauses = new ArrayList<Clause>();
				ruleIndex.put(r.getHead().getPredicate(), clauses);
			}
			clauses.add(r);

			for (Atom a : r.getBody()) {
				ArrayList<Clause> iclauses = iruleIndex.get(a.getPredicate());
				if (iclauses == null) {
					iclauses = new ArrayList<Clause>();
					iruleIndex.put(a.getPredicate(), iclauses);
				}
				iclauses.add(r);
			}
			
			if (!r.getHead().isFunctionFree()) {
				for (Term t : r.getHead().getTerms()) {
					if (t.isFunctionalTerm()) {
						Function f = ((FunctionalTerm)t).getFunction();
						Set<Clause> s = fIndex.get(f);
						if (s == null) {
							s = new HashSet<>();
							fIndex.put(f, s);
						}
						s.add(r);
					}
				}
			}
		}
		
	}

	
	public ArrayList<Clause> getClauses() {
		ArrayList<Clause> clauses = new ArrayList<Clause>();
		clauses.addAll(rules);
		
		return clauses;
	}

	public ArrayList<Clause> getClauses(Predicate pred) {
		ArrayList<Clause> res = ruleIndex.get(pred);
		if (res == null) {
			return new ArrayList<Clause>();
		} else {
			return res;
		}
	}	
	
	public ArrayList<Clause> getGraphDependencyProgram(Predicate pred) {
		Set<Predicate> preds = new HashSet<Predicate>();
		preds.add(pred);
		
		return getGraphDependencyProgram(preds);
	}
	
	public ArrayList<Clause> getGraphDependencyProgram(Set<Predicate> pred) {
		
		Set<Predicate> used = new HashSet<Predicate>();
		ArrayList<Predicate> queue = new ArrayList<Predicate>();
		Set<Predicate> reachablePredicates = new HashSet<Predicate>();
		
		int i = 0;
		queue.addAll(pred);
		used.addAll(pred);
		
		while (i < queue.size()){
			Predicate p = queue.get(i++);

			if (reachablePredicates.add(p)){
				for (Predicate newPred : getDirectReachablePredicates(p)) {
					if (used.add(newPred)) {
						queue.add(newPred);
					}
				}
			}
		}
		
		ArrayList<Clause> result = new ArrayList<Clause>();
		
		Set<Predicate> tmp = new HashSet<>(ruleIndex.keySet());
		tmp.removeAll(reachablePredicates);
		
		for (Predicate p : reachablePredicates) {
			ArrayList<Clause> cl = ruleIndex.get(p);
			if (cl != null) {
				result.addAll(cl);
			}
		}
		
		return result;
	}	
	
	private Set<Predicate> getDirectReachablePredicates(Predicate p) {
		Set<Predicate> res = new HashSet<Predicate>();
		
		ArrayList<Clause> cl = ruleIndex.get(p);
		if (cl != null) {
			for (Clause r : cl) {
				for (Atom t : r.getBody()) {
					res.add(t.getPredicate());
				}
			}
		}
		
		return res;
	}
	

	public String toString() {
		String  s = "";
		
//		for (int i = 0; i < rules.size(); i++) {
//			s += i + ": " + rules.get(i).toString() + "\n";
//		}
		
		int i = 0;
		for (Clause r : rules) {
			s += i++ + ": " + r.toString() + "\n";
		}
		
		return s;
	}
	
	
}
